[System.Reflection.Assembly]::LoadWithPartialName("System.Text.Json")
[System.Reflection.Assembly]::LoadWithPartialName("System.Net.Http")
[System.Reflection.Assembly]::LoadWithPartialName("System.Web")

[string]$popServerKey = ""

[string]$apiHost = "https://pay3.veritrans.co.jp"
[string]$apiPath = "/pop/v1/link-pay"

[System.Net.Http.HttpClient]$httpClient = [System.Net.Http.HttpClient]::new()

$DebugPreference = "Continue"

#region mail settings
class MailSettings {
    static [string] $smtpServerAddress = ""
    static [System.Boolean] $smtpEnableSsl = $false
    static [Int32] $smtpServerPort = 25
    static [System.Boolean] $smtpNeedCredentials = $false
    static [string] $smtpUser = ""
    static [string] $smtpPassword = ''
    static [string] $mailFrom = ''
    static [string] $mailSubject = "お支払いURLのご案内"
    static [string] $mailBody = @'
%NAME% 様

サンプルショップをご利用いただき、ありがとうございます。

注文内容
----------------------
取引ID：%ORDERID%
送料：%SHIPPING%円
合計金額：%AMOUNT%円
注文日：%ORDERDATE%
----------------------
%MAILINCLUDETEXT%

下記のURLから、決済手続きを開始してください。
%PAYMENTLINK%
'@
}
#endregion

function ReadCsv([System.Net.Http.HttpClient]$client, [string]$serverKey, [String]$path, [String]$encoding) {
    Write-Debug ([string]::Join($null, ((Get-Date), "[start processing]")))

    if ([String]::IsNullOrEmpty($encoding) -eq $true) {
        $encoding = "Default"
    }
    $orderCsv = Import-Csv -Encoding $encoding -Path $path

    $resultList = @()

    [UInt32]$line = 0

    $orderCsv | ForEach-Object {
        $line++
        Write-Debug ([string]::Join($null, ((Get-Date), "[csv line ${line}]", $_) ))

        #region AssignToVariable
        $order_id = $_.order_id
        $gross_amount = [string]::IsNullOrWhiteSpace($_.gross_amount) -eq $true ? $null : $_.gross_amount
        $mail_flag = $_.mail_flag
        $customer_name = [string]::IsNullOrWhiteSpace( $_.customer_name) -eq $true ? $null : $_.customer_name
        $customer_email_address = [string]::IsNullOrWhiteSpace($_.customer_email_address) -eq $true ? $null : $_.customer_email_address
        $shipping_amount = [string]::IsNullOrWhiteSpace($_.shipping_amount) -eq $true ? $null : $_.shipping_amount
        $dummy = [string]::IsNullOrWhiteSpace($_.dummy) -eq $true ? $null : $_.dummy
        $capture = [string]::IsNullOrWhiteSpace($_.capture) -eq $true ? $null : $_.capture
        $payment_key_expiry_duration = [string]::IsNullOrWhiteSpace($_.payment_key_expiry_duration) -eq $true ? $null : $_.payment_key_expiry_duration
        $push_url = [string]::IsNullOrWhiteSpace($_.push_url) -eq $true ? $null : $_.push_url
        $enable_payment_types = $_.enable_payment_types
        $mpi = [string]::IsNullOrWhiteSpace($_.mpi) -eq $true ? $null : $_.mpi
        $paynow_account_id = [string]::IsNullOrWhiteSpace($_.paynow_account_id) -eq $true ? $null : $_.paynow_account_id
        $save_card = [string]::IsNullOrWhiteSpace($_.save_card) -eq $true ? $null : $_.save_card
        $skip_card_selection = [string]::IsNullOrWhiteSpace($_.skip_card_selection) -eq $true ? $null : $_.skip_card_selection
        $cvs_types = $_.cvs_types
        $payment_expiry_duration = [string]::IsNullOrWhiteSpace($_.payment_expiry_duration) -eq $true ? $null : $_.payment_expiry_duration
        $payment_expiry_date = [string]::IsNullOrWhiteSpace($_.payment_expiry_date) -eq $true ? $null : $_.payment_expiry_date
        $contents = [string]::IsNullOrWhiteSpace($_.contents) -eq $true ? $null : $_.contents
        $contents_kana = [string]::IsNullOrWhiteSpace($_.contents_kana) -eq $true ? $null : $_.contents_kana
        $mail_include_text = [string]::IsNullOrWhiteSpace($_.mail_include_text) -eq $true ? $null : $_.mail_include_text
        $item_id_1 = [string]::IsNullOrWhiteSpace( $_.item_id_1) -eq $true ? $null : $_.item_id_1
        $item_name_1 = [string]::IsNullOrWhiteSpace($_.item_name_1) -eq $true ? $null : $_.item_name_1
        $item_price_1 = [string]::IsNullOrWhiteSpace($_.item_price_1) -eq $true ? $null : $_.item_price_1
        $item_quantity_1 = [string]::IsNullOrWhiteSpace($_.item_quantity_1) -eq $true ? $null : $_.item_quantity_1
        $item_id_2 = [string]::IsNullOrWhiteSpace($_.item_id_2) -eq $true ? $null : $_.item_id_2
        $item_name_2 = [string]::IsNullOrWhiteSpace($_.item_name_2) -eq $true ? $null : $_.item_name_2
        $item_price_2 = [string]::IsNullOrWhiteSpace($_.item_price_2) -eq $true ? $null : $_.item_price_2
        $item_quantity_2 = [string]::IsNullOrWhiteSpace($_.item_quantity_2) -eq $true ? $null : $_.item_quantity_2
        $item_id_3 = [string]::IsNullOrWhiteSpace($_.item_id_3) -eq $true ? $null : $_.item_id_3
        $item_name_3 = [string]::IsNullOrWhiteSpace($_.item_name_3) -eq $true ? $null : $_.item_name_3
        $item_price_3 = [string]::IsNullOrWhiteSpace($_.item_price_3) -eq $true ? $null : $_.item_price_3
        $item_quantity_3 = [string]::IsNullOrWhiteSpace($_.item_quantity_3) -eq $true ? $null : $_.item_quantity_3
        #endregion

        # result object
        $result = New-Object PSObject | Select-Object ORDER_ID, LINKPAY_URL_RESULT, URL, PAYMENT_KEY_EXPIRY_TIME, MESSAGE, MAIL_RESULT
        $result.ORDER_ID = $order_id
        
        [string]$json = $null

        try {
            $Error.Clear()
            # create request json
            $json = GetPostBody $order_id $gross_amount $mail_flag $customer_name $customer_email_address $shipping_amount `
                $dummy $capture $payment_key_expiry_duration $push_url $enable_payment_types $mpi $paynow_account_id `
                $save_card $skip_card_selection $cvs_types $payment_expiry_duration $payment_expiry_date $contents `
                $contents_kana $mail_include_text $item_id_1 $item_name_1 $item_price_1 $item_quantity_1 $item_id_2 `
                $item_name_2 $item_price_2 $item_quantity_2 $item_id_3 $item_name_3 $item_price_3 $item_quantity_3
            
            Write-Debug ([string]::Join($null, ((Get-Date), "[request json]", $json)))
        }
        catch [Exception] {
            $result.MESSAGE = $_.Exception.Message
            $resultList += $result

            Write-Debug ([string]::Join($null, ((Get-Date), "[create json error]", $_.Exception.Message)))

            return
        }

        try {
            $Error.Clear()
            # request to api
            $tuple = GetLinkPaymentUrl $client $json $serverKey
            $options = [System.Text.Json.JsonSerializerOptions]::new()
            $options.IgnoreNullValues = $true
            [ResponseEntity]$response = [System.Text.Json.JsonSerializer]::Deserialize($tuple.Item2, [ResponseEntity])

            Write-Debug ([string]::Join($null, ((Get-Date), "[response code]", $tuple.Item1)))
            
            if ($tuple.Item1 -eq 201) {
                # success
                $result.LINKPAY_URL_RESULT = $true
                $result.URL = $response.pay_link_url
                $result.PAYMENT_KEY_EXPIRY_TIME = $response.payment_key_expiry_time
            }
            else {
                # failure
                $result.LINKPAY_URL_RESULT = $false
                $result.MESSAGE = $response.message
            }
        }
        catch [Exception] {

            Write-Debug ([string]::Join($null, ((Get-Date), "[api call error]", $_.Exception.Message)))

            $result.LINKPAY_URL_RESULT = $false
            $result.MESSAGE = $_.Exception.Message
            $resultList += $result
            return
        }

        # if link generation succeeded and mail_flag = 1 or 3
        if ($result.LINKPAY_URL_RESULT -eq $true -and ($mail_flag -eq 1 -or $mail_flag -eq 3)) {

            Write-Debug ([string]::Join($null, ((Get-Date), "[send linkpay mail]")))

            [string]$orderDate = (Get-Date).ToString("yyyy/MM/dd")

            [string]$expireDate = $null
            [string]$expireTime = $null
            if ([String]::IsNullOrEmpty($result.PAYMENT_KEY_EXPIRY_TIME) -eq $false) {
                $expireDate = ([string]$result.PAYMENT_KEY_EXPIRY_TIME.Substring(0, 4)) + "/" + ([string]$result.PAYMENT_KEY_EXPIRY_TIME.Substring(4, 2)) + "/" + ([string]$result.PAYMENT_KEY_EXPIRY_TIME.Substring(6, 2))
                $expireTime = ([string]$result.PAYMENT_KEY_EXPIRY_TIME.Substring(8, 2)) + ":" + ([string]$result.PAYMENT_KEY_EXPIRY_TIME.Substring(10, 2)) + ":" + ([string]$result.PAYMENT_KEY_EXPIRY_TIME.Substring(12, 2))
            }

            try {
                $Error.Clear()
                # create mail body and subject
                [System.Tuple[[string], [string]]]$mail = CreateMailString ([MailSettings]::mailSubject) ([MailSettings]::mailBody) `
                    $order_id $result.URL $customer_name $gross_amount $shipping_amount $orderDate $expireDate `
                    $expireTime $mail_include_text            

                # send mail
                [System.Boolean]$mailResult = SendMail $mail.Item1 $mail.Item2 $customer_email_address $customer_name

                $result.MAIL_RESULT = $mailResult
            }
            catch {
                Write-Debug ([string]::Join($null, ((Get-Date), "[send mail error]", $_.Exception.Message)))

                $result.MESSAGE = $_.Exception.Message
                $result.MAIL_RESULT = $false
            }

        }

        $resultList += $result
    }

    $dir = [System.IO.Path]::GetDirectoryName($path)
    $outputFileName = [System.IO.Path]::GetFileNameWithoutExtension($path) + "_" + (Get-Date).ToString("yyMMdd_HHmmss") + "_result.csv"
    $outputPath = Join-Path $dir $outputFileName

    $resultList | Export-Csv $outputPath -Encoding $encoding -NoTypeInformation 

    Write-Debug ([string]::Join($null, ((Get-Date), "[end processing]")))
}

function GetPostBody {
    [OutputType([String])]
    param(
        $order_id, $gross_amount, $mail_flag, $customer_name, $customer_email_address, $shipping_amount,
        $dummy, $capture, $payment_key_expiry_duration, $push_url, $enable_payment_types, $mpi, $paynow_account_id,
        $save_card, $skip_card_selection, $cvs_types, $payment_expiry_duration, $payment_expiry_date, $contents, 
        $contents_kana, $mail_include_text, $item_id_1, $item_name_1, $item_price_1, $item_quantity_1, $item_id_2,
        $item_name_2, $item_price_2, $item_quantity_2, $item_id_3, $item_name_3, $item_price_3, $item_quantity_3
    )
        
    
    $e = [LinkPayEntity]::new($order_id, $gross_amount)

    if ($mail_flag -eq 2 -or $mail_flag -eq 3) {
        $mail = [LinkPayEmailEntity]::new()
        $mail.customer_email_address = [string]$customer_email_address
        $mail.customer_name = [string]$customer_name
        $e.email = $mail
    }
        
    if ($null -ne $shipping_amount) {
        $e.shipping_amount = [UInt32]$shipping_amount
    }

    if ($null -ne $dummy) {
        $e.dummy = [System.Boolean]::Parse($dummy)
    }
    if ($null -ne $capture) {
        $e.capture = [System.Boolean]::Parse($capture)
    }

    if ($null -ne $payment_key_expiry_duration) {
        $e.payment_key_expiry_duration = [UInt32]$payment_key_expiry_duration
    }
    
    if ($null -ne $push_url) {
        $e.push_url = [string]$push_url
    }
        
    if ($null -ne $enable_payment_types) {
        $bits = [Convert]::ToString($enable_payment_types, 2).PadLeft(3, '0')
        $enable = @();
        if ($bits.Substring($bits.Length - 1, 1) -eq "1") {
            $enable += "card"
        }
        if ($bits.Substring($bits.Length - 2, 1) -eq "1") {
            $enable += "cvs"
        }
        if ($bits.Substring($bits.Length - 3, 1) -eq "1") {
            $enable += "bank"
        }
        if ($enable.Length -gt 0) {
            $e.enabled_payment_types = $enable
        }
    }

    # card
    if ($null -ne $mpi -or $null -ne $paynow_account_id -or $null -ne $save_card -or $null -ne $skip_card_selection) {
        $card = [LinkPayCardEntity]::new()
        
        if ($null -ne $mpi) {
            $card.mpi = [System.Boolean]::Parse($mpi)
        }
        
        if ($null -ne $paynow_account_id) {
            $card.paynow_account_id = $paynow_account_id
        }
        if ($null -ne $save_card) {
            $card.save_card = $save_card
        }

        if ($null -ne $skip_card_selection) {
            $card.skip_card_selection = [System.Boolean]::Parse($skip_card_selection)
        }

        $e.card = $card
    }

    # cvs
    if ([string]::IsNullOrWhiteSpace($cvs_types) -ne $true) {
        $cvsArray = @();
        $bits = [Convert]::ToString($cvs_types, 2).PadLeft(3, '0')
        if ($bits.Substring($bits.Length - 1, 1) -eq "1") {
            $cvs = [LinkPayCvsEntity]::new()
            $cvs.payment_sub_type = "sej"
            if ($null -ne $payment_expiry_duration) {
                $cvs.payment_expiry_duration = $payment_expiry_duration
            }
            if ($null -ne $payment_expiry_date) {
                $cvs.payment_expiry_date = $payment_expiry_date
            }
            $cvsArray += $cvs
        }
        if ($bits.Substring($bits.Length - 2, 1) -eq "1") {
            $cvs = [LinkPayCvsEntity]::new()
            $cvs.payment_sub_type = "econ"
            if ($null -ne $payment_expiry_duration) {
                $cvs.payment_expiry_duration = $payment_expiry_duration
            }
            if ($null -ne $payment_expiry_date) {
                $cvs.payment_expiry_date = $payment_expiry_date
            }
            $cvsArray += $cvs
        }
        if ($bits.Substring($bits.Length - 3, 1) -eq "1") {
            $cvs = [LinkPayCvsEntity]::new()
            $cvs.payment_sub_type = "other"
            if ($null -ne $payment_expiry_duration) {
                $cvs.payment_expiry_duration = $payment_expiry_duration
            }
            if ($null -ne $payment_expiry_date) {
                $cvs.payment_expiry_date = $payment_expiry_date
            }
            $cvsArray += $cvs
        }
        $e.cvs = $cvsArray
    }

    # bank
    if ($null -ne $contents -or $null -ne $contents_kana) {
        $bank = [LinkPayBankEntity]::new()
        $bankSubTypea = [LinkPayBankPaymentSubTypeEntity]::new()
        $bankSubTypea.payment_sub_type = "atm"
        if ($null -ne $payment_expiry_duration) {
            $bankSubTypea.payment_expiry_duration = $payment_expiry_duration
        }
        if ($null -ne $payment_expiry_date) {
            $bankSubTypea.payment_expiry_date = $payment_expiry_date
        }
        $bank.contents = $contents
        $bank.contents_kana = $contents_kana
        $e.bank = $bank
    }

    [LinkPayItemEntity[]]$items = @()
    if ($null -ne $item_id_1 -or $null -ne $item_name_1 -or $null -ne $item_price_1 -or $null -ne $item_quantity_1) {
        $item = [LinkPayItemEntity]::new($item_id_1, $item_name_1, $item_price_1, $item_quantity_1)
        $items += $item
    }
    if ($null -ne $item_id_2 -or $null -ne $item_name_2 -or $null -ne $item_price_2 -or $null -ne $item_quantity_2) {
        $item = [LinkPayItemEntity]::new($item_id_2, $item_name_2, $item_price_2, $item_quantity_2)
        $items += $item
    }
    if ($null -ne $item_id_3 -or $null -ne $item_name_3 -or $null -ne $item_price_3 -or $null -ne $item_quantity_3) {
        $item = [LinkPayItemEntity]::new($item_id_3, $item_name_3, $item_price_3, $item_quantity_3)
        $items += $item
    }
    if ($items.Length -gt 0 ) {
        $e.items = $items
    }

    $options = [System.Text.Json.JsonSerializerOptions]::new()
    $options.IgnoreNullValues = $true
    $json = [System.Text.Json.JsonSerializer]::Serialize($e, $options)
    return $json
}

function GetLinkPaymentUrl {
    [OutputType([System.Tuple[[System.Net.HttpStatusCode], [string]]])]
    param([System.Net.Http.HttpClient]$client, [string]$json, [string]$serverKey) 

    if ([String]::IsNullOrEmpty($serverKey) -eq $true) {
        throw "POP Server Key is null or empty."
    }

    $content = [System.Net.Http.StringContent]::new($json, [System.Text.Encoding]::UTF8, "application/json")

    $bytes = ([System.Text.Encoding]::UTF8).GetBytes($serverKey + ":")
    $base64encoded = [Convert]::ToBase64String($bytes)

    $task = $null
    
    $client.DefaultRequestHeaders.Authorization = [System.Net.Http.Headers.AuthenticationHeaderValue]::new("Basic", $base64encoded)
    $url = $apiHost + $apiPath
    $Error.Clear()
    $task = $client.PostAsync($url, $content)
    $task.Wait()

    $response = $task.Result

    $json = $response.Content.ReadAsStringAsync().Result
    Write-Debug ([string]::Join($null, ((Get-Date), "[response json]", $json) ))

    return [System.Tuple[[System.Net.HttpStatusCode], [string]]]::new($response.StatusCode, $json)
}

#region request json entity
class LinkPayEntity {
    [string]$order_id
    [System.Nullable[UInt32]]$gross_amount
    [System.Nullable[UInt32]]$shipping_amount
    [System.Nullable[System.Boolean]]$dummy
    [System.Nullable[System.Boolean]]$capture
    [System.Nullable[UInt32]]$payment_key_expiry_duration
    [string]$push_url
    [string[]]$enabled_payment_types
    [LinkPayCardEntity]$card
    [LinkPayCvsEntity[]]$cvs
    [LinkPayBankEntity]$bank
    [LinkPayEmailEntity]$email
    [LinkPayItemEntity[]]$items

    LinkPayEntity([string]$order_id, [System.Nullable[UInt32]]$gross_amount) {
        $this.order_id = $order_id
        $this.gross_amount = $gross_amount
    }

}

class LinkPayCardEntity {
    [System.Nullable[System.Boolean]]$mpi
    [string]$paynow_account_id
    [string]$save_card # always|optional
    [System.Nullable[System.Boolean]]$skip_card_selection
}

class LinkPayCvsEntity {
    [string]$payment_sub_type # sej|econ|other
    [System.Nullable[UInt16]]$payment_expiry_duration # Max value -> sej:150, econ:60, other:365
    [string]$payment_expiry_date # YYYYMMDD
}

class LinkPayBankEntity {
    [LinkPayBankPaymentSubTypeEntity]$payment_sub_type
    [string]$contents
    [string]$contents_kana
}

class LinkPayBankPaymentSubTypeEntity {
    [string]$payment_sub_type # atm
    [System.Nullable[UInt16]]$payment_expiry_duration # Max value -> 60
    [string]$payment_expiry_date # YYYYMMDD
}

class LinkPayEmailEntity {
    [string]$customer_name
    [string]$customer_email_address
}

class LinkPayItemEntity {
    [string]$id
    [string]$name
    [System.Nullable[Int32]]$price
    [System.Nullable[UInt32]]$quantity # Greater than 0

    LinkPayItemEntity( [string]$id,
        [string]$name,
        [System.Nullable[Int32]]$price,
        [System.Nullable[UInt32]]$quantity) {
        $this.id = $id
        $this.name = $name
        $this.price = $price
        $this.quantity = $quantity
    }
}
#endregion

#region response json entity
class ResponseEntity {
    [string]$payment_key
    [string]$result_code
    [string]$status
    [string]$message
    [string]$payment_key_expiry_time
    [string]$pay_link_url
}
#endregion

function SendMail {
    [OutputType([System.Boolean])]
    param($subject, $body, $to, $customerName) 
    
    if ([String]::IsNullOrWhiteSpace([MailSettings]::smtpServerAddress) -eq $false) {
        $client = New-Object System.Net.Mail.SmtpClient([MailSettings]::smtpServerAddress, [MailSettings]::smtpServerPort)
        $client.EnableSsl = [MailSettings]::smtpEnableSsl
        if ($smtpNeedCredentials -eq $true) {
            $client.Credentials = New-Object System.Net.NetworkCredential([MailSettings]::smtpUser, [MailSettings]::smtpPassword)
        }
        $mailAddress = [System.Net.Mail.MailAddress]::new($to, $customerName)
        $message = [System.Net.Mail.MailMessage]::new([MailSettings]::mailFrom, $mailAddress, $subject, $body)
        $client.Send($message)
        return $true
    }
    return $false
    
}

function CreateMailString {
    [OutputType([System.Tuple[[string], [string]]])]
    param([string]$subject, [string]$body, [string]$orderId, [string]$paymentLink, [string]$name, 
        [System.Nullable[UInt32]]$amount, [System.Nullable[UInt32]]$shipping, [string]$orderDate,
        [string]$expireDate, [string]$expireTime, [string]$mailIncludeText)

    $subject = $subject.Replace("%ORDERID%", $orderId)
    $subject = $subject.Replace("%PAYMENTLINK%", $paymentLink)
    $subject = $subject.Replace("%NAME%", $name)
    $subject = $subject.Replace("%AMOUNT%", $amount.ToString("#,#"))
    if ([string]::IsNullOrWhiteSpace($shipping) -ne $true) {
        $subject = $subject.Replace("%SHIPPING%", $shipping.ToString("#,#"))
    }
    else { $subject = $subject.Replace("%SHIPPING%", $null) }

    $subject = $subject.Replace("%ORDERDATE%", $orderDate)
    $subject = $subject.Replace("%EXPIREDATE%", $expireDate)
    $subject = $subject.Replace("%EXPIRETIME%", $expireTime)

    $body = $body.Replace("%ORDERID%", $orderId)
    $body = $body.Replace("%PAYMENTLINK%", $paymentLink)
    $body = $body.Replace("%NAME%", $name)
    $body = $body.Replace("%AMOUNT%", $amount.ToString("#,#"))
    if ([string]::IsNullOrWhiteSpace($shipping) -ne $true) {
        $body = $body.Replace("%SHIPPING%", $shipping.ToString("#,#"))
    }
    else { $body = $body.Replace("%SHIPPING%", $null) }
    $body = $body.Replace("%ORDERDATE%", $orderDate)
    $body = $body.Replace("%EXPIREDATE%", $expireDate)
    $body = $body.Replace("%EXPIRETIME%", $expireTime)
    $body = $body.Replace("%MAILINCLUDETEXT%", $mailIncludeText)

    return [System.Tuple[[string], [string]]]::new($subject, $body)
}


if ($Args.Count -eq 1) {
    ReadCsv $httpClient $popServerKey $Args[0]
}
elseif ($Args.Count -eq 2) {
    ReadCsv $httpClient $popServerKey $Args[0] $Args[1]
}
else {
    Write-Host "`n・利用方法`n"
    Write-Host "第1引数でファイル名を指定してください。"
    Write-Host "> .\csv2poplinkpay.ps1 .\TestData1.csv`n"
    Write-Host "第2引数で読み込むCSVファイルの文字コードを指定することができます。"
    Write-Host "* 現在のデフォルト文字コード:" ([System.Text.Encoding]::Default.EncodingName)
    Write-Host "未指定の場合はシステムのデフォルト文字コードが利用されます。"
    Write-Host "> .\csv2poplinkpay.ps1 .\TestData1.csv SJIS"
}
